pub mod test;

use rgb::RGB8;
use crate::lib8::Hsv;

use crate::events::EventBus;
use crate::geometry::*;
use crate::render::{Shader, Surface, Surfaces};
use crate::task::Task;
use crate::lib8::{trig::{sin8, cos8}, noise::inoise8, IntoRgb8};

#[derive(Debug)]
pub struct IdleTask<T: Surface> {
    solid: T,
    surface: T,
    shimmer: T,
}

#[derive(Debug)]
struct SolidShader {}

impl Shader for SolidShader {
    fn draw(&self, _coords: &VirtualCoordinates, frame: usize) -> RGB8 {
        Hsv::new((frame % 255) as u8, 255, sin8((frame % 64) as u8)).into_rgb8()
    }
}

#[derive(Debug)]
struct ShimmerShader {}

impl Shader for ShimmerShader {
    fn draw(&self, coords: &VirtualCoordinates, frame: usize) -> RGB8 {
        Hsv::new(((coords.x as usize).wrapping_add(frame / 3) % 255) as u8, coords.y, sin8(frame).max(75).min(64)).into_rgb8()
    }
}

#[derive(Debug)]
struct ThinkingShader {}

impl Shader for ThinkingShader {
    fn draw(&self, coords: &VirtualCoordinates, frame: usize) -> RGB8 {
        //let noise_x = sin8(sin8((frame % 255) as u8).wrapping_add(coords.x));
        //let noise_y = cos8(cos8((frame % 255) as u8).wrapping_add(coords.y));
        let offset_x = sin8(frame.wrapping_add(coords.x as usize));
        let offset_y = cos8(frame.wrapping_add(coords.y as usize));
        let noise_x = offset_x / 2;
        let noise_y = offset_y / 2;
        //let noise_x = coords.x.wrapping_add(offset_x);
        //let noise_y = coords.y.wrapping_add(offset_y);
        Hsv::new(
            inoise8(offset_x as i16, offset_y as i16),
            128_u8.saturating_add(inoise8(noise_y as i16, noise_x as i16)),
            255
        ).into_rgb8()
    }
}

impl<T: Surface> IdleTask<T> {
    pub fn new<S: Surfaces<Surface = T>>(surfaces: &mut S) -> Self {
        IdleTask {
            solid: surfaces.new_surface(Rectangle::everything()).unwrap(),
            surface: surfaces.new_surface(Rectangle::everything()).unwrap(),
            shimmer: surfaces.new_surface(Rectangle::everything()).unwrap(),
        }
    }
}

impl<T: Surface> Task for IdleTask<T> {
    fn name(&self) -> &'static str { "Idle" }

    fn start(&mut self, _bus: &mut EventBus) {
        self.solid.set_shader(SolidShader {});
        self.surface.set_shader(ThinkingShader { });
        self.shimmer.set_shader(ShimmerShader { });

        self.solid.set_opacity(64);
        self.surface.set_opacity(128);
        self.shimmer.set_opacity(64);
    }

    fn stop(&mut self, _bus: &mut EventBus) {
        self.solid.clear_shader();
        self.surface.clear_shader();
        self.shimmer.clear_shader();
    }
}