use core::fmt::Debug;
use std::{collections::VecDeque, fmt::{Display, Result}, sync::{Arc, Mutex}};

use crate::{properties::{Properties, PropertyID, Variant}, property_namespace};

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum Event {
    ReadyToRock,
    Tick,
    StartThing(&'static str),
    StopThing(&'static str),
    PropertyChange(PropertyID, Variant)
}

impl Display for Event {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result {
        match self {
            Self::ReadyToRock => f.write_str("ReadyToRock"),
            Self::Tick => f.write_str("Tick"),
            Self::StartThing(name) => write!(f, "Start {}", *name),
            Self::StopThing(name) => write!(f, "Stop {}", *name),
            Self::PropertyChange(id, value) => write!(f, "{} -> {:?}", id, value)
        }
    }
}

impl Event {
    pub const fn new_tick() -> Self {
        Event::Tick
    }

    pub fn new_property_change<T>(key: PropertyID, data: T) -> Self where Variant: From<T> {
        Event::PropertyChange(key, Variant::from(data))
    }

    pub const fn new_ready_to_rock() -> Self {
        Event::ReadyToRock
    }

    pub const fn new_start_thing(name: &'static str) -> Self {
        Event::StartThing(name)
    }

    pub const fn new_stop_thing(name: &'static str) -> Self {
        Event::StopThing(name)
    }
}

#[derive(Clone)]
pub struct EventBus {
    pending: Arc<Mutex<VecDeque<Event>>>,
    props: Arc<Mutex<Properties>>
}

impl EventBus {
    pub fn new() -> Self {
        EventBus {
            pending: Arc::new(Mutex::new(VecDeque::with_capacity(32))),
            props: Arc::new(Mutex::new(Properties::new()))
        }
    }

    pub fn next(&mut self) -> Event {
        let next = self.pending.lock().unwrap().pop_front().unwrap_or(Event::new_tick());

        match next {
            Event::PropertyChange(key, value) => {
                if self.props.lock().unwrap().set(key, value.clone()) {
                    log::trace!("prop-update key={:?} value={:?}", key, value);
                    Event::PropertyChange(key, value)
                } else {
                    Event::new_tick()
                }
            },
            _ => next
        }
    }

    pub fn push(&mut self, event: Event) {
        self.pending.lock().unwrap().push_back(event);
    }

    pub fn properties(&self) -> std::sync::MutexGuard<'_, Properties> {
        self.props.lock().unwrap()
    }

    pub fn set_property<K: Into<PropertyID>, V: Into<Variant>>(&mut self, key: K, value: V) {
        self.push(Event::new_property_change(key.into(), value.into()));
    }
}

property_namespace!(
    System,
    NetworkOnline => false,
    IP => "",
    Gateway => "",
    TimeSync => false
);