use std::{collections::LinkedList, thread::JoinHandle};

use esp_idf_svc::mqtt::client::{EspMqttClient, MqttClientConfiguration};
use serde_json::{json, Value};

use crate::{events::{Event, EventBus, System as SystemNS}, prop_id, properties::{PropertyID, Variant}, render::props::Output as OutputNS, scenes::props::Scenes as SceneNS, task::Task};
use crate::platform::props::Board as BoardNS;
use paste::paste;

struct HADevice {
    prefix: String,
    unique_id: String
}

impl HADevice {
    fn new(component: &str, chip_id: u64, name: &str) -> Self {
        HADevice {
            // eg: homeassistant/sensor/0BADCOFFEE/fps
            unique_id: format!("{:X}-{}", chip_id, name),
            prefix: format!("homeassistant/{}/renderbug-rs/{:X}-{}", component, chip_id, name)
        }
    }

    fn topic(&self, name: &str) -> String {
        format!("{}/{}", self.prefix, name)
    }

    fn registration(&self) -> Value {
        json!({
            "~": self.prefix,
            "stat_t": "~/state",
            "unique_id": self.unique_id,
            "dev": {
                "name": "Renderbug-rs ESP32",
                "mdl": "Renderbug-rs ESP32",
                "sw": "",
                "mf": "Phong Robotics",
                "ids": [self.unique_id]
            }
        })
    }
}

struct HAScene {
    prefix: String,
    unique_id: String
}

impl HAScene {
    fn new(chip_id: u64, name: &str) -> Self {
        HAScene {
            // eg: homeassistant/sensor/0BADCOFFEE/fps
            unique_id: format!("{:X}-{}", chip_id, name),
            prefix: format!("homeassistant/scene/renderbug-rs/{:X}-{}", chip_id, name)
        }
    }

    fn topic(&self, name: &str) -> String {
        format!("{}/{}", self.prefix, name)
    }

    fn registration(&self) -> Value {
        json!({
            "~": self.prefix,
            "stat_t": "~/state",
            "cmd_t": "~/command",
            "unique_id": self.unique_id,
            "payload_on": "on",
            "dev": {
                "name": "Renderbug-rs ESP32",
                "mdl": "Renderbug-rs ESP32",
                "sw": "",
                "mf": "Phong Robotics",
                "ids": [self.unique_id]
            }
        })
    }
}

pub struct MqttTask {
    client: Option<EspMqttClient<'static>>,
    conn_thread: Option<JoinHandle<()>>,
    fps_sensor: Option<HADevice>
}

impl MqttTask {
    pub fn new() -> Self {
        MqttTask {
            conn_thread: None,
            client: None,
            fps_sensor: None
        }
    }

    fn start_mqtt(&mut self, bus: &EventBus) {
        log::info!("Starting MQTT");
        let chip_id: u64 = bus.properties().get(BoardNS::ChipID).unwrap().into();

        self.fps_sensor = Some(HADevice::new("sensor", chip_id, "output-fps"));

        let (client, mut conn) = EspMqttClient::new(
            "mqtt://10.0.0.2:1883",
            &MqttClientConfiguration {
                client_id: Some(&format!("{:X}", chip_id)),
                ..Default::default()
            }
        ).unwrap();
        log::info!("Connected!");

        self.conn_thread = Some(std::thread::Builder::new()
            .stack_size(6000)
            .spawn(move || {
                conn.next().unwrap();
            }).unwrap());
        self.client = Some(client);
    }
}

impl Task for MqttTask {
    fn start(&mut self, bus: &mut EventBus) {
        bus.set_property(props::MQTT::Online, false);

        let chip_id = bus.properties().get(BoardNS::ChipID).unwrap().into();
        self.fps_sensor = Some(HADevice::new("sensor", chip_id, "fps"));
    }

    fn on_property_change(&mut self, key: PropertyID, value: &Variant, bus: &mut EventBus) {
        match (key, value) {
            (prop_id!(SystemNS::NetworkOnline), Variant::Boolean(true)) => {
                log::info!("Registering with MQTT");

                let chip_id = bus.properties().get(BoardNS::ChipID).unwrap().into();

                self.start_mqtt(bus);

                if let Some(ref mut client) = self.client {
                    if let Some(ref sensor) = self.fps_sensor {
                        client.enqueue(
                            &sensor.topic("config"),
                            esp_idf_svc::mqtt::client::QoS::AtLeastOnce,
                            false,
                            sensor.registration().to_string().as_bytes()
                        ).unwrap();
                    }

                    let scenes: Vec<Variant> = bus.properties().get(SceneNS::All).unwrap().into();
                    for scene in scenes.iter() {
                        let scene_name: String = scene.clone().into();
                        let scene_device = HAScene::new(chip_id, scene_name.as_str());
                        client.enqueue(
                            &scene_device.topic("config"),
                            esp_idf_svc::mqtt::client::QoS::AtLeastOnce,
                            false,
                            scene_device.registration().to_string().as_bytes()
                        ).unwrap();
                    }
                    log::info!("MQTT should be online!");

                    bus.set_property(props::MQTT::Online, true);
                }
            },
            (prop_id!(OutputNS::FPS), Variant::UInt(fps)) => {
                if let Some(ref mut client) = self.client {
                    if let Some(ref sensor) = self.fps_sensor {
                        client.enqueue(
                            &sensor.topic("state"),
                            esp_idf_svc::mqtt::client::QoS::AtLeastOnce,
                            false,
                            json!(fps).to_string().as_bytes()
                        ).unwrap();
                    }
                }
            },
            _ => ()
        }
    }
}

pub mod props {
    use crate::property_namespace;

    property_namespace!(
        MQTT,
        Online => true
    );
}